SOFTWARE_VERSION_NUMBER = "2.6.0"
DEVICE_TYPE = "STELLA-Q2"
# STELLA-Q2 spectrometer instrument
# NASA open source software license
# Paul Mirel 2024

# set the sampling period = reciprocal of the sampling cadence
#               seconds + ( minutes ) + ( hours )
sample_interval_s = 1.0 + ( 0 * 60 ) + ( 0 * 3600 )
burst_count = 3
logplot = True #False

#Gain 0: 1x
#Gain 1: 3.7x
#Gain 2: 16x (calibrated at this value)
#Gain 3: 64x
triad_gain = 2

# integration time integer set from 0 to 255
# integration time will be = 2.8ms * [integration cycles + 1]
# 59 cycles = 166ms (calibrated at this value)
# max int time = 717ms
triad_int_cycles = 59 #59


### TBD ###
# later: repair the header when triad is not present.
# later: set clock to GPS time if available. do this once on the first good fix, if latitude is not None
# later: add option for +$5 128x128 display (3D address), 128x32 is 3C
# case: add arrow and REC markings
# case: add location for GPS module


## Manual: ##
## starts up in continuous recording
## click the button to pause recording (response is a bit slow ~1s)
## click again to take a data point burst, as many as you want.
## Set the number of measurements per burst on line 13.
## press-and-hold (2sec) to turn on the lamps,
## press-and-hold again to turn off the lamps.

### TBD ###
# use the GPS time to set the real time clock (TBD)

# set the welcome display cycle wait time
welcome_wait_s = 2

import gc
gc.collect()
start_mem_free_kB = gc.mem_free()/1000
print("start memory free {0:.2f} kB".format( start_mem_free_kB ))
import time
import os
import microcontroller
import board
import digitalio
import rtc
import neopixel
import storage
import sdcardio
import busio
import math
import adafruit_max1704x
from adafruit_pcf8523 import pcf8523
import displayio
import terminalio
import adafruit_displayio_ssd1306
from adafruit_display_text import label
import vectorio
from i2c_button import I2C_Button
import adafruit_gps
import AS7265X_sparkfun
from AS7265X_sparkfun import AS7265X
print( "mem free after imports = {} kB, {} %".format(int(gc.mem_free()/1000), int(100*(gc.mem_free()/1000)/start_mem_free_kB )) )
if burst_count < 1:
    burst_count = 1

def main():
    display = False
    displayio.release_displays()
    UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
    print("unique identifier (UID) : {0}".format( UID ))
    onboard_blue_led = initialize_led( board.LED )
    onboard_neopixel = initialize_neopixel( board.NEOPIXEL )
    try:
        vfs, sdcard = initialize_sd_card_storage()
        storage.mount(vfs, "/sd")
    except:
        print( "SD card fail: missing or full" )
        vfs = False
        sdcard = False


    success_color = ( 255, 0, 0 ) #Green
    fault_color = ( 102, 255, 0 ) #Orange
    OFF = ( 0, 0, 0 )

    i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )

    try:
        display_bus = displayio.I2CDisplay( i2c_bus, device_address=0x3c )
        display = adafruit_displayio_ssd1306.SSD1306( display_bus, width=128, height=32 )
        display_group = displayio.Group()
        print( "found 128x32 display" )
    except:
        display = False
        display_group = False

        # 0x68 -- PCF8523 real time clock
    null_time = time.struct_time(( 2020,  01,   01,   00,  00,  00,   0,   -1,    -1 ))
    try:
        hardware_clock = pcf8523.PCF8523(i2c_bus)
        print( "found hardware_clock" )
        clock_battery_low = hardware_clock.battery_low
        if clock_battery_low:
            print( "    clock battery is low. replace clock battery" )
        else:
            print( "    clock battery is OK" )
        timenow = hardware_clock.datetime
        if timenow.tm_wday not in range ( 0, 7 ):
            hardware_clock.datetime = null_time
        timenow = hardware_clock.datetime
    except (ValueError, NameError) as err:
        print( "hardware clock fail: {}".format(err))
        hardware_clock = False
        clock_battery_low = True
        timenow = null_time
    try:
        system_clock = rtc.RTC()
        system_clock.datetime = hardware_clock.datetime
        print( "    system clock synchronized to pcf8523_real_time_clock" )
    except:
        print( "failed to set system clock" )
    DAYS = { 0:"Sunday", 1:"Monday", 2:"Tuesday",
        3:"Wednesday", 4:"Thursday", 5:"Friday", 6:"Saturday" }

    # create or update the batch number
    batch_number = 0
    if hardware_clock:
        batch_number = update_batch( timenow )
    print( "batch number == {}".format( batch_number ))

    # 0x6f -- i2c_button
    try:
        i2c_button = I2C_Button( i2c_bus )
        print( "found i2c button" )
        #print("    firmware version", i2c_button.version)
        #print("    i2c button debounce ms", i2c_button.debounce_ms)
        i2c_button.clear()
    except:
        i2c_button = False

    battery_monitor = initialize_battery_monitor( i2c_bus )
    battery_monitor.report()
    battery_monitor.read()

    if display:
        # date time
        if hardware_clock:
            timenow = hardware_clock.datetime
        else:
            timenow = null_time
        iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format(
            timenow.tm_year, timenow.tm_mon, timenow.tm_mday,
            timenow.tm_hour, timenow.tm_min, timenow.tm_sec )
        weekday = DAYS[ timenow.tm_wday ]
        display.root_group = display_group

        date_group = displayio.Group(scale=2, x=3, y=8)
        text = "{}-{:02}-{:02}".format(timenow.tm_year, timenow.tm_mon, timenow.tm_mday)
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        date_group.append(text_area) # Subgroup for text scaling

        display_group.append(date_group)

        time_group = displayio.Group(scale=1, x=3, y=28)
        text = "{:02}:{:02} UTC".format( timenow.tm_hour, timenow.tm_min )
        batch_number_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        time_group.append(batch_number_area) # Subgroup for text scaling
        display_group.append(time_group)
        print( "show date and time on display" )
        time.sleep( welcome_wait_s )

    triad_spectrometer = initialize_triad_spectrometer( i2c_bus )
    triad_spectrometer.found()
    triad_spectrometer.report()
    triad_spectrometer.read()

    ## triad_spectrometer check and set gain ##
    #Gain 0: 1x
    #Gain 1: 3.7x
    #Gain 2: 16x
    #Gain 3: 64x
    print(triad_spectrometer.swob._gain)
    triad_spectrometer.swob.set_gain( triad_gain )
    confirm_gain = triad_spectrometer.swob._gain
    if confirm_gain < 1:
        gain_ratio = 1
    elif confirm_gain == 1:
        gain_ratio = 3.7
    elif confirm_gain == 2:
        gain_ratio = 16
    elif confirm_gain == 3:
        gain_ratio = 64
    print("triad spectrometer gain set to {} == {}X".format(confirm_gain, gain_ratio))

    ## triad_spectrometer check and set integration time ##
    #print( triad_spectrometer.swob.virtual_read_register(INTERGRATION_TIME))
    triad_spectrometer.swob.set_integration_cycles(triad_int_cycles)
    int_time_ms = round((2.8*(triad_int_cycles+1)),0)
    print("triad spectrometer integration time set to {} == {}ms".format(triad_int_cycles, int_time_ms))


    triad_spectrometer.lamps_on()
    time.sleep(1)
    triad_spectrometer.lamps_off()
    triad_spectrometer.read()
    print( "WL, cal, cal.uncty, counts, chip, chip_temp_C" )
    for item in triad_spectrometer.bands_sorted:
        print("{}, {}, {}, {}, {}, {}".format( item,
            triad_spectrometer.dict_cal[item],
            triad_spectrometer.dict_cal[item] * triad_spectrometer.uncert_percent/100,
            triad_spectrometer.dict_counts[item], triad_spectrometer.dict_chip_n[item],
            triad_spectrometer.chip_temps[triad_spectrometer.dict_chip_n[item]] ))

    if display:
        # UID/ SD
        display_group.pop()
        display_group.pop()
        text_group = displayio.Group(scale=2, x=0, y=15)
        text = "U:" + str(UID)
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)
        batch_number_group = displayio.Group(scale=2, x=80, y=15)
        if vfs:
            text = "B:"+str( batch_number )
        else:
            text = "NoSD"
        batch_number_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        batch_number_group.append(batch_number_area) # Subgroup for text scaling
        display_group.append(batch_number_group)
        print( "show UID and batch on display" )
        time.sleep( welcome_wait_s )

    if display:
        # battery
        display_group.pop()
        display_group.pop()
        text_group = displayio.Group(scale=2, x=0, y=15)
        text = "battery"
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)

        battery_monitor.read()
        battery_percent_group = displayio.Group(scale=2, x=90, y=15)
        text = "{}%".format( int( battery_monitor.percentage ))
        battery_percent_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        battery_percent_group.append(battery_percent_area) # Subgroup for text scaling
        display_group.append(battery_percent_group)
        print( "show battery percent on display" )
        time.sleep( welcome_wait_s )

    if display:
        # gain
        display_group.pop()
        display_group.pop()
        text_group = displayio.Group(scale=2, x=15, y=15)
        text = "Gain: "
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)

        gain_group = displayio.Group(scale=2, x=80, y=15)
        text = "{}X".format( gain_ratio )
        gain_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        gain_group.append(gain_area) # Subgroup for text scaling
        display_group.append(gain_group)
        time.sleep( welcome_wait_s )

    if display:
        # integration time
        display_group.pop()
        display_group.pop()
        text_group = displayio.Group(scale=2, x=0, y=15)
        text = "IntT: "
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)

        integration_group = displayio.Group(scale=2, x=70, y=15)
        text = "{}ms".format( int(int_time_ms) )
        integration_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        integration_group.append(integration_area) # Subgroup for text scaling
        display_group.append(integration_group)
        time.sleep( welcome_wait_s )

    gps = initialize_gps( i2c_bus )
    gps.report()
    gps.send_start_commands()
    gps.read()

    if display:
        display_group.pop()
        display_group.pop()
        text_group = displayio.Group(scale=2, x=0, y=15)
        text = "Intv:"
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)

        sample_interval_group = displayio.Group(scale=2, x=60, y=15)
        text = "{}s".format( sample_interval_s )
        if sample_interval_s > 59:
            text = "{}m".format( round( sample_interval_s / 60, 1 ))
        if sample_interval_s > 3600:
            text = "{}h".format( round( sample_interval_s / 3600, 1 ))
        interval_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        sample_interval_group.append(interval_area) # Subgroup for text scaling
        display_group.append(sample_interval_group)
        time.sleep( welcome_wait_s )
        display_group.pop()
        display_group.pop()

        text_group = displayio.Group(scale=2, x=0, y=15)
        text = "burst:"
        text_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        text_group.append(text_area) # Subgroup for text scaling
        display_group.append(text_group)
        burst_count_group = displayio.Group(scale=2, x=90, y=15)
        text = "{}".format( burst_count )
        burst_count_area = label.Label(terminalio.FONT, text=text, color=0xFFFFFF)
        burst_count_group.append(burst_count_area) # Subgroup for text scaling
        display_group.append(burst_count_group)

        time.sleep( welcome_wait_s )
        display_group.pop()
        display_group.pop()



    #build the header
    header = "iso8601_utc, UID, batch, mmn, dec.hour, Gain.X, intT.ms"
    if triad_spectrometer.swob:
        header += ", "
        header += triad_spectrometer.header()
    if gps.swob:
        header += ", "
        header += gps.header()
    if battery_monitor.swob:
        header += ", "
        header += battery_monitor.header()
    header += "\n"
    previous_header = "None"

    # generate filename for today
    fod = ("SQ2_data_{}{:02}{:02}.csv".format(timenow.tm_year, timenow.tm_mon,timenow.tm_mday))
    previous_finu = "placeholder for filename in use"
    previous_header = "placeholder header"
    finu = fod
    # look to see what the previous filename_in_use was
    try:
        with open( "/sd/fn.txt", "r" ) as fn:
            try:
                previous_finu = fn.readline().rstrip()
                finu = previous_finu
            except ValueError:
                previous_finu = "no previous filename"
    except OSError:
            print( "fn.txt file not found, creating new fn.txt file" )
            try:
                with open ( "/sd/fn.txt", "w" ) as fn:
                    fn.write(fod)
            except:
                print( "unable to create fn.txt file")

    try:
        with open( ( "/sd/{}".format( finu )), "r" ) as fn:
            previous_header = fn.readline().rstrip()
            previous_header += "\n"
    except:
        print( "error opening datafile" )

    if False: #previous_finu != finu:
        # create new datafile named finu. write header
        try:
            # create a new data file, and write the header in it
            with open( ( "/sd/{}".format( finu )), "w" ) as f:
                f.write(header)
                #f.write("\n")
            print( "created new data file for today" )
            recording = True
        except OSError as err:
            print( "error opening datafile, {}".format( err ))
            recording = False

    #print( len( previous_header))
    #print( previous_header )
    #print()
    #print( len( header ))
    print( header )


    if header == previous_header:
        print( "headers match" )
    else:
        finu = ("SQ2_data_{}{:02}{:02}-{}.csv".format(timenow.tm_year, timenow.tm_mon,timenow.tm_mday,batch_number))
        print( "finu = {}".format( finu ))
        try:
            # create a new data file, and write the header in it
            with open( ( "/sd/{}".format( finu )), "w" ) as f:
                f.write( header )
            print( "created new data file for the new configuration" )
            recording = True
        except OSError as err:
            print( "error opening datafile, {}".format( err ))
            recording = False
        try:
            with open( "/sd/fn.txt", "w" ) as fn:
                fn.write( finu )
        except:
            print( "fn.txt file not available for update" )

    interval_start_time_s = time.monotonic()
    loop_count = 0
    operational = True
    recording = True
    one_shot = False
    lamps_on = False
    try:
        graph_bar, graph_bar_x, batch_number_label_text_area, polygon = create_graph_screen( display_group, triad_spectrometer.bands_sorted )
        # create log label and backing block
        log_tag_group = displayio.Group(scale=1, x=55, y=27)
        text = "log"
        log_tag_area = label.Label(terminalio.FONT, text=text, color=0x000000)
        log_tag_group.append(log_tag_area) # Subgroup for text scaling
        display_group.append(log_tag_group)
        if logplot:
            log_tag_group.hidden = False
        else:
            log_tag_group.hidden = True

        gps_tag_group = displayio.Group(scale=1, x=90, y=27)
        text = "nof"
        gps_tag_area = label.Label(terminalio.FONT, text=text, color=0x000000)
        gps_tag_group.append(gps_tag_area) # Subgroup for text scaling
        display_group.append(gps_tag_group)
        if gps.swob:
            gps_tag_group.hidden = False
        else:
            gps_tag_group.hidden = True

        while operational:
            loop_begin_time_s = time.monotonic()
            loop_count += 1
            if loop_count > 1023:
                loop_count = 0
            gc.collect()
            if i2c_button:
                button_values = i2c_button.status
                if button_values[ 1 ] or button_values[2]:
                    print( "button clicked" )
                    recording = False
                    one_shot = True
                    batch_number = update_batch( hardware_clock.datetime )
                    i2c_button.led_bright = 16
                if button_values[ 2 ] and not button_values[1]:
                    time.sleep(0.1)
                    hold_time = 0
                    press_hold = True
                    start_hold_time = time.monotonic()
                    while button_values[ 2 ] and hold_time <= 2.2:
                        button_values = i2c_button.status
                        hold_time = time.monotonic() - start_hold_time
                        time.sleep(0.1)
                    if hold_time > 2:
                        lamps_on = not lamps_on
                        if lamps_on:
                            print( "turn on the lamps")
                            triad_spectrometer.lamps_on()
                        else:
                            print( "turn off the lamps")
                            triad_spectrometer.lamps_off()
                i2c_button.clear()


            # check if it is time to read a new datapoint. If not yet, then skip this section
            if time.monotonic() > interval_start_time_s + sample_interval_s:
                gc.collect()
                if display:
                    if recording:
                        if vfs:
                            batch_number_label_text_area.text = "batch:" + str( batch_number )
                        else:
                            batch_number_label_text_area.text = "NoSDcard"
                    else:
                        if vfs:
                            batch_number_label_text_area.text = "lastB:" + str( batch_number )
                        else:
                            batch_number_label_text_area.text = "NoSDcard"



                interval_start_time_s = time.monotonic()
                mem_free_kB = gc.mem_free()/1000
                print( "\nmemory free == {} kB, {:.1f} %".format( mem_free_kB, 100 * mem_free_kB/start_mem_free_kB ))
                check_loop_count = loop_count

                triad_spectrometer.read()
                battery_monitor.read()
                gps.read()
                if triad_spectrometer.swob:
                    gps_tag_area.color = 0x000000
                else:
                    gps_tag_area.color = 0xFFFFFF
                if gps.latitude is not None:
                    gps_tag_area.text = "FIX"
                else:
                    gps_tag_area.text = "nof"

                spectral_data_sorted = []
                for item in triad_spectrometer.bands_sorted:
                    if logplot:
                        #show log tag and base rectangle
                        if triad_spectrometer.dict_cal[item] > 0:
                            spectral_data_sorted.append(math.log(triad_spectrometer.dict_cal[item],10))
                        else:
                            spectral_data_sorted.append(0)
                    else:
                        #hide log tag and base rectangle
                        spectral_data_sorted.append(triad_spectrometer.dict_cal[item])
                    #print(item, triad_spectrometer.dict_cal[item],
                    #    triad_spectrometer.dict_cal[item] * triad_spectrometer.uncert_percent/100,
                    #    triad_spectrometer.dict_counts[item], triad_spectrometer.dict_chip_n[item],
                    #    triad_spectrometer.chip_temps[triad_spectrometer.dict_chip_n[item]] )
                print( spectral_data_sorted )

                gc.collect()
                graph_data( spectral_data_sorted, graph_bar, graph_bar_x, polygon )
                mmn = 0
                while mmn < burst_count:
                    mmn += 1
                    triad_spectrometer.read()
                    if hardware_clock:
                        timenow = hardware_clock.datetime
                    else:
                        timenow = time.struct_time(( 2020,  01,   01,   00,  00,  00,   0,   -1,    -1 ))
                    iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format(
                        timenow.tm_year, timenow.tm_mon, timenow.tm_mday,
                        timenow.tm_hour, timenow.tm_min, timenow.tm_sec )
                    decimal_hour = timestamp_to_decimal_hour( timenow )
                    if recording or one_shot:
                        try:
                            with open( "/sd/{}".format( finu ), "a" ) as f:
                                if onboard_neopixel:
                                    onboard_neopixel.fill ( success_color )
                                for item in triad_spectrometer.bands_sorted:
                                    f.write( "{}, {}, ".format( iso8601_utc, UID) )
                                    f.write( "{}, {}, {}, {}, {}, ".format( batch_number, mmn, decimal_hour, gain_ratio, int(int_time_ms)))
                                    f.write( "{}, {}, {}, {}, {}, {}".format(
                                        item,
                                        triad_spectrometer.dict_cal[item],
                                        triad_spectrometer.dict_cal[item] * triad_spectrometer.uncert_percent/100,
                                        triad_spectrometer.dict_counts[item],
                                        triad_spectrometer.dict_chip_n[item],
                                        triad_spectrometer.chip_temps[triad_spectrometer.dict_chip_n[item]]
                                        ))
                                    if gps.swob:
                                        f.write( ", " )
                                        f.write( gps.log())
                                    if battery_monitor.swob:
                                        f.write( ", " )
                                        f.write( battery_monitor.log())
                                    f.write("\n")
                                f.close()
                            if onboard_neopixel:
                                onboard_neopixel.fill( OFF )
                        except OSError as err:
                            print( "\nError: sd card fail: {:}".format(err) )
                            print( "*** The card may be missing, or you may need to create a folder named sd in the root directory of CIRCUITPY ***" )
                            if onboard_neopixel:
                                onboard_neopixel.fill( fault_color ) #  fault_color to show error: likely no SD card present, or SD card full.
                                time.sleep( 1 )
                                onboard_neopixel.fill( OFF )
                            recording = False
                if one_shot == True:
                    i2c_button.led_bright = 0
                one_shot = False
            time.sleep(0.1)
            loop_time = (time.monotonic() - loop_begin_time_s)

    finally:  # clean up the busses when ctrl-c'ing out of the loop
        if sdcard:
            sdcard.deinit()
            print( "sd card deinitialized" )
        displayio.release_displays()
        print( "displayio displays released" )
        i2c_bus.deinit()
        print( "i2c_bus deinitialized" )

class Device: #parent class
    def __init__(self, name = None, pn = None, address = None, swob = None):
        self.name = name
        self.swob = swob
        self.pn = pn
        self.address = address
    def report(self):
        if self.swob is not None:
            print("report:", self.name, self.pn, hex(self.address), "found" )

def initialize_triad_spectrometer( i2c_bus ):
    triad_spectrometer = Null_Triad_Spectrometer()
    try:
        triad_spectrometer = as7265x_Triad_Spectrometer( i2c_bus )
    except:
        pass
    return triad_spectrometer

class as7265x_Triad_Spectrometer( Device ):
    def __init__( self, com_bus ):
        super().__init__(name = "triad_spectrometer", pn = "as7256x", address = 0x00, swob = AS7265X( com_bus ))
        self.swob.disable_indicator()
        self.swob.set_measurement_mode(AS7265X_sparkfun.MEASUREMENT_MODE_6CHAN_CONTINUOUS)
        self.bands = 610, 680, 730, 760, 810, 860, 560, 585, 645, 705, 900, 940, 410, 435, 460, 485, 510, 535
        self.chip_n = 1,   1,   1,   1,   1,   1,   2,   2,   2,   2,   2,   2,   3,   3,   3,   3,   3,   3
        self.dict_chip_n = {key:value for key, value in zip(self.bands, self.chip_n)}
        self.bands_sorted = sorted( self.bands )
        self.uncert_percent = 12
    def found(self):
        print("found", self.pn, self.swob)
    def read(self):
        self.chip_temps = {1:self.swob.get_temperature(1), 2:self.swob.get_temperature(2), 3:self.swob.get_temperature(3)}
        self.data_counts = self.swob.get_value(0)
        self.dict_counts = {key:value for key, value in zip(self.bands, self.data_counts)}
        self.data_cal = self.swob.get_value(1)
        self.dict_cal = {key:value for key, value in zip(self.bands, self.data_cal)}
    def list_channels():
        return self.bands_sorted
    def header( self ):
        return "WL.nm, irrad.uW/(cm^2), irrad.uncty.uW/(cm^2), counts, chip_num, chip_temp_C"
    def log( self ):
        print( self.data_counts )
        print( self.data_cal )
        #self.irradiance[ch] = self.data[ch]/self.tsis_cal_counts_per_irradiance[ch]
        #self.irradiance[ch] = self.data[ch]/self.steno_cal_counts_per_irradiance[ch]
        #return "{}, {}, {}, {}".format( self.center_wavelengths_nm[ch], self.data[ch],
        #    self.irradiance[ch], self.irradiance[ch]*self.calibration_error )
    def printlog(self,ch):
        print( self.log(ch) )
    def lamps_on(self):
        print( "turn on the lamps")
        self.swob.enable_bulb(0)   # white
        self.swob.enable_bulb(1)   # NIR
        self.swob.enable_bulb(2)   # UV
    def lamps_off(self):
        print( "turn off the lamps")
        self.swob.disable_bulb(0)   # white
        self.swob.disable_bulb(1)   # NIR
        self.swob.disable_bulb(2)   # UV

class Null_Triad_Spectrometer():
    def __init__( self ):
        self.swob = None
        self.bands_sorted = [0,0]
        self.dict_chip_n = [0,0]
        self.chip_temps = [0,0]
        self.dict_cal = {0:10}
        self.dict_counts = {0:10}
        self.uncert_percent = 10
    def found(self):
        pass
    def read(self):
        pass
    def log(self):
        pass
    def report(self):
        pass
    def printlog(self):
        pass
    def blink(self, duration):
        pass
    def header(self):
        pass
    def lamps_on(self):
        pass
    def lamps_off(self):
        pass

def initialize_gps( i2c_bus ):
    gps = Null_GPS()
    try:
        gps = pa1010d_GPS( i2c_bus )
    except:
        pass
    return gps

class pa1010d_GPS( Device ):
    def __init__( self, com_bus ):
        super().__init__(name = "gps", pn = "pa1010d", address = 0x10, swob = adafruit_gps.GPS_GtopI2C( com_bus, debug=False ))
    def found(self):
        print("found", self.pn, self.swob)
    def send_start_commands(self):
        self.swob.send_command(b"PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0") #set data output configuration
        self.swob.send_command(b"PMTK220,1000") #set update interval to 1000 ms
    def read(self):
        self.swob.update()
        self.latitude = self.swob.latitude
        self.lat_deg = self.swob.latitude_degrees
        self.lat_min = self.swob.latitude_minutes
        self.longitude = self.swob.longitude
        self.long_deg = self.swob.longitude_degrees
        self.long_min = self.swob.longitude_minutes
        self.altitude = self.swob.altitude_m
    def header(self):
        return( "lat.deg.N, long.deg.W, alt.m" )
    def log(self):
        return "{}, {}, {}".format( self.latitude, self.longitude, self.altitude )
    def printlog(self):
        print( self.log())

class Null_GPS():
    def __init__( self ):
        self.swob = None
        self.latitude = None
    def found(self):
        pass
    def read(self):
        pass
    def log(self):
        pass
    def report(self):
        pass
    def send_start_commands(self):
        pass
    def printlog(self):
        pass
    def header(self):
        pass

def initialize_battery_monitor( i2c_bus ):
    battery_monitor = Null_Battery_Monitor()
    try:
        battery_monitor = max1704x_Battery_Monitor( i2c_bus )
    except:
        pass
    return battery_monitor

class max1704x_Battery_Monitor( Device ): #child class ( parent class ):
    def __init__( self, com_bus ):
        super().__init__(name = "battery_monitor", pn = "max1704x", address = 0x36, swob = adafruit_max1704x.MAX17048( com_bus ))
    def found(self):
        print("found", self.pn, self.swob)
    def read(self):
        self.voltage = self.swob.cell_voltage
        self.percentage = self.swob.cell_percent
    def header(self):
        return "bat.V, bat.%"
    def log(self):
        return "{}, {}".format( self.voltage, self.percentage )
    def printlog(self):
        print( self.log())

class Null_Battery_Monitor():
    def __init__( self ):
        self.swob = None
    def found(self):
        pass
    def read(self):
        pass
    def log(self):
        pass
    def report(self):
        pass
    def printlog(self):
        pass
    def header(self):
        pass

def graph_data( spectral_data_sorted, graph_bar, graph_bar_x, polygon ):
    if spectral_data_sorted:
        y_zero_pixels = 31
        y_upper_pixels = 17
        graph_points = [(2, 112), (2, 104)]
        y_span_pixels = y_zero_pixels - y_upper_pixels
        irradiances = spectral_data_sorted
        irrad_min = min( irradiances )
        irrad_max = max( irradiances )
        irrad_span = irrad_max - irrad_min
        if irrad_span < 1:
            irrad_span = 1
        for count in range (len( irradiances )):
            irrad_value = irradiances[ count ]
            irrad_height = irrad_value - irrad_min
            irrad_normalized_height = irrad_height/ irrad_span
            irrad_bar_height_pixels = int(y_span_pixels * irrad_normalized_height)
            irrad_drop_height_pixels = int(y_span_pixels - irrad_bar_height_pixels)
            irrad_y_top_pixel = int(y_upper_pixels + irrad_drop_height_pixels)
            graph_bar[count].y = irrad_y_top_pixel
            graph_bar[count].height = 1
            point = (graph_bar_x[count]+1, irrad_y_top_pixel)
            graph_points.append(point)
        graph_points.append((126, 104))
        graph_points.append((126, 112))
        polygon.points = graph_points
        #del y_zero_pixels, y_upper_pixels, graph_points, y_span_pixels, irradiances, spectral_data_sorted
        #del irrad_min, irrad_max, irrad_span, count, irrad_value, irrad_height
        #del irrad_normalized_height, irrad_bar_height_pixels, irrad_drop_height_pixels
        #del irrad_y_top_pixel, point

def create_graph_screen( display_group, bands_sorted ):
    if display_group is not False:
        palette = displayio.Palette(1)
        palette[0] = 0xFFFFFF
        points2 = [ (2, 32), (2, 30), (126, 30), (126, 32)]
        polygon = vectorio.Polygon(pixel_shader=palette, points=points2, x=0, y=0)
        display_group.append( polygon )
        bar_color = 0xFFFFFF
        bar = displayio.Palette(1)
        bar[0] = bar_color
        bar_width = 1
        bar_default_height = 1
        graph_bar_x = (0, 6, 12, 18, 24, 29, 35, 41, 47, 55, 63, 69, 75, 82, 93, 105, 114, 124)
        graph_bar=[]
        for count in range( len( bands_sorted)):
            graph_bar.append( vectorio.Rectangle(pixel_shader=bar, width=bar_width, height=bar_default_height, x=graph_bar_x[count], y=106))
            display_group.append( graph_bar[count] )

        wavelength_label_group = displayio.Group( scale=1, x=4, y=9)
        wavelength_label_text = "410nm          940nm"
        wavelength_label_text_area = label.Label( terminalio.FONT, text=wavelength_label_text, color=0xFFFFFF )
        wavelength_label_group.append( wavelength_label_text_area )
        display_group.append( wavelength_label_group )

        batch_number_label_group = displayio.Group( scale=1, x=38, y=9)
        batch_number_label_text = "batch:"
        batch_number_label_text_area = label.Label( terminalio.FONT, text=batch_number_label_text, color=0xFFFFFF )
        batch_number_label_group.append( batch_number_label_text_area )
        display_group.append( batch_number_label_group )

        print( "initialized graph screen" )
        return ( graph_bar, graph_bar_x, batch_number_label_text_area, polygon )
    else:
        print( "graph screen initialization failed" )
        return False, False, False, False, False, False, False, False, False, False, False, False



def update_batch( timestamp ):
    gc.collect()
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            except ValueError:
                previous_batch_number = 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0

    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
            b.write("\n")
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def request_firmware_report( gps_sensor ):
    gps_sensor.send_command(b"PMTK605")  # request firmware version
    data = gps_sensor.read(32)  # read up to 32 bytes
        # print(data)  # this is a bytearray type
    if data is not None:
        # convert bytearray to string
        data_string = "".join([chr(b) for b in data])
        #print( "gps firmware report = ", end='')
        report = data_string[:4]
        #print(report, end="")
    else:
        report = None
        print( "no data on firmware" )
    #print("\n")
    return report

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
        print( "initialized i2c_bus" )
    except ValueError as err:
        i2c_bus = False
        print( "i2c bus fail: {} -- press reset button, or power off to restart".format( err ))
    return i2c_bus

def initialize_sd_card_storage( ):
    try:
        cs = board.SD_CS
        sd_card_spi = busio.SPI(board.SD_SCK, MOSI=board.SD_MOSI, MISO=board.SD_MISO)
        sdcard = sdcardio.SDCard(sd_card_spi, cs)
        vfs = storage.VfsFat(sdcard)
        print( "sdcardio success" )
    except( OSError, ValueError ) as err:
        print( "No SD card found, or card is full: {}".format(err) )
        vfs = False
        sdcard = False
        onboard_neopixel.fill( 255,0,0 )
    return vfs, sdcard

def initialize_neopixel( pin ):
    num_pixels = 1
    ORDER = neopixel.RGB
    neopixel_instance = neopixel.NeoPixel( pin, num_pixels, brightness=0.3, auto_write=True, pixel_order=ORDER )
    return neopixel_instance

def initialize_led( pin ):
    LED = digitalio.DigitalInOut( pin )
    LED.direction = digitalio.Direction.OUTPUT
    count = 4
    interval = 0.1
    LED.value = True
    time.sleep(interval)
    LED.value = False
    return LED

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def memory_check( message ):
    gc.collect()
    mem_free_kB = gc.mem_free()/1000
    print( "{} memory free: {} kB, {} %".format( message, int(mem_free_kB), int((100* (mem_free_kB)/start_mem_free_kB ))))

def stall():
    print("intentionally stalled, press return to continue")
    input_string = False
    while input_string == False:
        input_string = input().strip()

gc.collect()
main()
